@@ -199,7 +199,8 @@ class BinRegexps:
         return False    
 
     def is_needed(self, f):
-        return (self.is_just_copy(f) or self.is_need_patch(f)) and not self.is_need_exclude(f)
+        return (self.is_just_copy(f) or self.is_need_patch(f)) 
+        #and not self.is_need_exclude(f)
     
     pass
 
@@ -255,6 +256,7 @@ class PackagesSpec:
     terra:  list 
     exclude_prefix: list 
     exclude_suffix: list
+    remove_from_download: list
 
     def __post_init__(self):    
         '''
@@ -467,7 +469,7 @@ class TerrariumAssembler:
         May be here we will can catch output and hunt for heizenbugs
         '''    
         print(scmd)
-        os.system(scmd)
+        return os.system(scmd)
 
     def packages2list(self, pl):
         pl_ = []
@@ -665,8 +667,8 @@ pipenv run python3 -m pip freeze > {target_dir_}/{build_name}-pip-freeze.txt
             wtf333 = 1
 
 
-        if self.br.is_need_exclude(f):
-            return False
+        # if self.br.is_need_exclude(f):
+        #     return False
 
         if self.br.is_needed(f):
             return True
@@ -882,6 +884,8 @@ pipenv run python3 -m pip freeze > {target_dir_}/{build_name}-pip-freeze.txt
         return res_
 
     def add(self, what, to_=None, recursive=True):
+        if 'java-11' in what:
+            dfsfsfdsf=1
         try:
             if not to_:
                 to_ = what
@@ -893,9 +897,11 @@ pipenv run python3 -m pip freeze > {target_dir_}/{build_name}-pip-freeze.txt
             # ar.add(f)
             if os.path.isdir(what):
                 # copy_tree(what, os.path.join(root_dir, to_))
-                if not os.path.exists(os.path.join(self.root_dir, to_)):
-                    shutil.copytree(what, os.path.join(self.root_dir, to_), symlinks=True, copy_function=self.mycopy)
-                    #, exist_ok=True)
+                # Какого хуя!!!!!!!!!!
+                # if not os.path.exists(os.path.join(self.root_dir, to_)):
+                #     shutil.copytree(what, os.path.join(self.root_dir, to_), symlinks=True, copy_function=self.mycopy)
+                #     #, exist_ok=True)
+                pass
             else:
                 self.mycopy(what, os.path.join(self.root_dir, to_))
             pass
@@ -986,7 +992,12 @@ pipenv run python3 -m pip freeze > {target_dir_}/{build_name}-pip-freeze.txt
         lines2 = []
         in_src = os.path.relpath(self.src_dir, start=self.curdir)
         # lines.add("rm -rf %s " % in_src)
-        lines.append("mkdir -p %s " % in_src)
+        lines.append(f"""
+mkdir -p tmp/snaphots-src
+snapshotdir=$(date +"tmp/snaphots-src/snapshot-src-before-%Y-%m-%d-%H-%M-%S")
+mv in/src $snapshotdir
+mkdir -p {in_src}
+""")
         already_checkouted = set()
         for td_ in self.pp.projects() + self.spec.templates_dirs:
             git_url, git_branch, path_to_dir_, _ = self.explode_pp_node(td_)
@@ -1003,6 +1014,8 @@ pushd %(newpath)s
 git checkout %(git_branch)s
 git config core.fileMode false
 git config core.autocrlf input
+git lfs install 
+git lfs pull
 popd
 ''' % vars()
                 lines.append(scmd)
@@ -1090,7 +1103,7 @@ fi
         os.chdir(self.curdir)
 
         os.chdir(os.path.join('in', 'src', 'pip'))
-        scmd = f'''{self.root_dir}/ebin/python3 setup.py install  '''
+        scmd = f'''{self.root_dir}/ebin/python3 setup.py install --single-version-externally-managed --root / '''
         os.system(scmd)
 
         os.chdir(self.curdir)
@@ -1116,7 +1129,7 @@ fi
             '''
 
         os.chdir(self.curdir)
-        os.system(scmd)
+        self.cmd(scmd)
         if self.tvars.fc_version == '32':
             os.system(f"rm -f {root_dir}/local/lib/python3.8/site-packages/typing.*")
 
@@ -1166,14 +1179,15 @@ fi
                 # scmd = "%(root_dir)s/ebin/python3 setup.py install --single-version-externally-managed  %(release_mod)s --root / --force   " % vars()
                 self.cmd(f"{root_dir}/ebin/python3 setup.py install --single-version-externally-managed  {release_mod} --root / --force  ")  #--no-deps
 
-                os.chdir(setup_path)
-                for reqs_ in glob.glob(f'**/package.json', recursive=True):
-                    os.chdir(setup_path)
-                    dir_ = os.path.split(reqs_)[0]
-                    if dir_:
-                        os.chdir(dir_)
-                    os.system(f"yarn install ")
-                    os.system(f"yarn build ")
+                # os.chdir(setup_path)
+                # for reqs_ in glob.glob(f'**/package.json', recursive=True):
+                #     if not 'node_modules' in reqs_:
+                #         os.chdir(setup_path)
+                #         dir_ = os.path.split(reqs_)[0]
+                #         if dir_:
+                #             os.chdir(dir_)
+                #         os.system(f"yarn install ")
+                #         os.system(f"yarn build ")
 
 
         if self.tvars.fc_version == '32':
@@ -1192,7 +1206,7 @@ fi
     
         for rp_ in self.ps.repos or []:
             if rp_.endswith('.rpm'):
-                lines.append(f'sudo dnf install {rp_} -y ')
+                lines.append(f'sudo dnf install --nogpgcheck {rp_} -y ')
             else:
                 lines.append(f'sudo yum-config-manager --add-repo {rp_} -y ')
             pass
@@ -1227,6 +1241,10 @@ fi
         scmd = 'dnf download --skip-broken --downloaddir "%(in_bin)s/src-rpms" --arch=x86_64 --arch=noarch  --source %(packages)s -y ' % vars()
         lines_src.append(scmd)
 
+        for pack_ in self.ps.remove_from_download or []:
+            scmd = f'rm -f {in_bin}/rpms/{pack_}* '
+            lines.append(scmd)
+
         # for package in self.dependencies(pls_, local=False) + purls_:
         #     # потом написать идемпотентность, проверки на установленность, пока пусть долго, по одному ставит
         #     scmd = 'dnf download --downloaddir "%(in_bin)s/rpms" --arch=x86_64 "%(package)s" -y ' % vars()
@@ -1241,7 +1259,7 @@ fi
         shfilename = "02-install-rpms"    
         ilines = [
 """
-sudo dnf install --skip-broken %(in_bin)s/rpms/*.rpm -y --allowerasing
+sudo dnf install --nogpgcheck --skip-broken %(in_bin)s/rpms/*.rpm -y --allowerasing
 """ % vars()
         ]
         self.lines2sh("02-install-rpms", ilines, "install-rpms")    
@@ -1267,7 +1285,7 @@ sudo dnf install --skip-broken %(in_bin)s/rpms/*.rpm -y --allowerasing
             # scmd = "pushd %s" % (path_to_dir)
             # lines.append(scmd)
             scmd = f"""
-pipenv run sh -c "pushd {path_to_dir};python3 setup.py bdist_wheel -d {relwheelpath};popd" 
+pipenv run sh -c "pushd {path_to_dir};python3 setup.py clean --all;python3 setup.py bdist_wheel -d {relwheelpath};popd" 
 """
             lines.append(scmd)
             pass
@@ -1282,6 +1300,7 @@ pipenv run sh -c "pushd {path_to_dir};python3 setup.py bdist_wheel -d {relwheelp
         lines = []
         scmd = f'''
 pipenv --rm
+rm -f Pipfile*
 pipenv install --python {self.tvars.python_version_1}.{self.tvars.python_version_2}
 pipenv run python3 -m pip install ./in/bin/extwheel/*.whl --find-links="{ext_whl_path}"  --force-reinstall --ignore-installed  --no-cache-dir --no-index 
 ''' 
@@ -1324,8 +1343,8 @@ pipenv run python3 -m pip install ./in/bin/ourwheel/*.whl ./in/bin/extwheel/*.wh
 
 
         if terra:
-            projects += self.pp.terra.projects
-            pip_targets += self.pp.terra.pip
+            projects += self.pp.terra.projects or []
+            pip_targets += self.pp.terra.pip or []
         else:
             projects += self.pp.projects()
             pip_targets += self.pp.pip()
@@ -1347,7 +1366,16 @@ pipenv run python3 -m pip install ./in/bin/ourwheel/*.whl ./in/bin/extwheel/*.wh
 
             reqs_path = 'requirements.txt'
             for reqs_ in glob.glob(f'**/{reqs_path}', recursive=True):
-                pip_reqs.append(os.path.join(os.path.relpath(setup_path, start=self.curdir), reqs_))
+                if 'django-q' in setup_path:
+                    rewqwerew=1
+                with open(reqs_, 'r', encoding='utf-8') as lf:
+                    reqs__ = lf.read()
+                    if not '--hash=sha256' in reqs__:
+                        # very strange requirements.txt, we cannot download it.
+                        pip_reqs.append(os.path.join(os.path.relpath(setup_path, start=self.curdir), reqs_))
+                    else:
+                        print(f'«--hash=sha256» in {setup_path}/{reqs_}')
+                        
             pass
 
         return pip_targets, pip_reqs
@@ -1409,32 +1437,30 @@ rm -f *.tar.*
 
         trace_file = None
         try:
-            trace_file = spec.tests.tracefile
+            trace_file_glob = spec.tests.tracefile
         except:
             print('You should specify tests→tracefile')    
             return
 
         used_files = set()
+        for trace_file in glob.glob(trace_file_glob):
+            re_file = re.compile(r'''.*\(.*"(?P<filename>[^"]+)".*''')
+            for line in open(trace_file, 'r', encoding='utf-8').readlines():
+                m_ = re_file.match(line)
+                if m_:
+                    fname = m_.group('filename')
+                    # Heuristic to process strace files from Vagrant virtualboxes
+                    fname = fname.replace('/vagrant', self.curdir)
+                    # Heuristic to process strace files from remote VM, mounted by sshmnt
+                    fname = re.sub(fr'''/mnt/.*{lastdirs}''', abs_path_to_out_dir, fname)
+                    if os.path.isabs(fname):
+                        fname = os.path.abspath(fname)
+                        if fname.startswith(abs_path_to_out_dir):
+                            if os.path.islink(fname):
+                                link_ = os.readlink(fname)
+                                fname = os.path.abspath(os.path.join(os.path.split(fname)[0], link_))
+                            used_files.add(os.path.abspath(fname))
 
-        re_file = re.compile(r'''.*\(.*"(?P<filename>[^"]+)".*''')
-        for line in open(trace_file, 'r', encoding='utf-8').readlines():
-            m_ = re_file.match(line)
-            if m_:
-                fname = m_.group('filename')
-                # Heuristic to process strace files from Vagrant virtualboxes
-                fname = fname.replace('/vagrant', self.curdir)
-                # Heuristic to process strace files from remote VM, mounted by sshmnt
-                if '/mnt/local/home/stas/projects/deploy-for-audit/linux_distro/out/ebin/../pbin/libtorch_cpu.so' in fname:
-                   wtrr=1
-                fname = re.sub(fr'''/mnt/.*{lastdirs}''', abs_path_to_out_dir, fname)
-                if os.path.isabs(fname):
-                    fname = os.path.abspath(fname)
-                    if fname.startswith(abs_path_to_out_dir):
-                        if os.path.islink(fname):
-                            link_ = os.readlink(fname)
-                            fname = os.path.abspath(os.path.join(os.path.split(fname)[0], link_))
-                        used_files.add(os.path.abspath(fname))
-        # print("\n".join(sorted(used_files)))
         existing_files = {}
         for dirpath, dirnames, filenames in os.walk(abs_path_to_out_dir):
             for filename in filenames:
@@ -1450,8 +1476,32 @@ rm -f *.tar.*
         top10 = sorted(existing_files.items(), key=lambda x: -x[1])[:1000]
         print("Analyse first:") 
         for f, s in top10:
-            f_ = f.replace(abs_path_to_out_dir, '    - .*')
-            print(f'{f_} # \t {s}')
+            rel_f = f.replace(abs_path_to_out_dir, '')
+            unban_ = False
+            for unban in [#'/usr/lib64/python3.9','/lib64/python3.9'
+                        ]:
+                if unban in rel_f:
+                    unban_ = True
+                    breakpoint
+            if unban_ :
+                continue        
+            # f_ = re.escape(rel_f)
+            f_ = rel_f
+            for rex_ in [
+                r"\.so\.[\d]+\.[\d]+\.[\d]+", #act\.so\.4\.0\.1 → act\.so\.\d\.\d\.\d 
+                r"\.so\.[\d]+\.[\d]+", 
+                r"\.so\.[\d]+", 
+                r"[\d]+\.[\d]+\.so", 
+                r"[\d]+\.so", 
+                r"-[\d]+\.[\d]+-[\d]+", 
+            ]:
+                def replacement_f(mobj):
+                    return rex_
+                f_ = re.sub(rex_, replacement_f, f_)
+            if f_ == rel_f:    
+               f_ = re.escape(rel_f)
+
+            print(f'      - .*{f_} # \t {s} \t {rel_f}')
 
         # print("\n".join([f'{f}: \t {s}' for f,s in top10]))
 
@@ -1467,6 +1517,8 @@ rm -f *.tar.*
         banned_start = ['tmp']
         banned_mid = ['/out/', '/wtf/', '/.vagrant/', '/.git/']
 
+        # there are regularly some files unaccessable for reading.
+        self.cmd('sudo chmod a+r /usr/lib/cups -R')
         self.cmd('systemd-tmpfiles --remove dnf.conf')
 
         def filter_(tarinfo):
@@ -1494,6 +1546,25 @@ rm -f *.tar.*
         tar.add(self.curdir, recursive=True, filter=filter_)
         tar.close()    
 
+    def remove_exclusions(self):
+        '''
+        Postprocessing, removing not needed files after installing python modules, etc
+        '''
+        from pathlib import Path
+        for path in Path(self.root_dir).rglob('*'):
+            rp_ = str(path.absolute())
+            if '/mildata/baselines_model.ckpt' in rp_:
+                sfdsfds =1
+            if self.br.is_need_exclude(rp_):
+                    path.unlink(missing_ok=True)
+
+        # killing broken links    
+        for path in Path(self.root_dir).rglob('*'):
+            rp_ = str(path.absolute())
+            if os.path.islink(rp_) and not os.path.exists(rp_):
+                    path.unlink(missing_ok=True)
+
+        pass            
 
     def process(self):
         '''
@@ -1505,6 +1576,11 @@ rm -f *.tar.*
         root_dir = self.root_dir 
         
         def install_templates(root_dir, args):
+            if 'prepare_install' in self.spec:
+                for scmd in self.spec.prepare_install.strip().split('\n'):
+                    self.cmd(scmd)
+
+
             if 'copy_folders' in self.spec:
                 for it_ in self.spec.copy_folders:
                     pass
@@ -1585,6 +1661,9 @@ rm -f *.tar.*
             libc_path = "/lib64/" + _findSoname_ldconfig('c')
             shutil.copy2(libc_path, f'{root_dir}/lib64/libc.so', follow_symlinks=True)
 
+            ebin_ = os.path.join(root_dir, 'ebin')
+            self.cmd(f'chmod a+x {ebin_}/*')
+
             print("Install templates takes")
             t.toc()
             pass
@@ -1622,7 +1701,6 @@ rm -f *.tar.*
         specfile_ = self.args.specfile
         self.lines2sh("50-pack", [
             '''
-sudo chmod a+rx /usr/lib/cups -R           
 #terrarium_assembler --stage-pack=./out "%(specfile_)s" --stage-make-isoexe
 terrarium_assembler --stage-pack=./out "%(specfile_)s" 
             ''' % vars()])
@@ -1646,23 +1724,31 @@ terrarium_assembler --debug --stage-pack=./out-debug "%(specfile_)s" --stage-mak
             ''' % vars()])
 
         root_dir = 'out'
+
+
         if self.args.stage_pack:
             root_dir = self.root_dir = expandpath(args.stage_pack)
+            # self.remove_exclusions()
 
             # install_templates(root_dir, args)
 
-            packages_to_deploy = self.ps.terra
+            packages_to_deploy = []
+            pips_to_deploy = []
             if self.args.debug:
                 packages_to_deploy += self.ps.terra + self.ps.build
+                pips_to_deploy = self.pp.pip()
+            else:
+                packages_to_deploy = self.ps.terra
+                pips_to_deploy = self.pp.terra.pip or []
 
 
-            fs_ = self.generate_file_list_from_pips(self.pp.pip())
+            fs_ = self.generate_file_list_from_pips(pips_to_deploy)
             file_list = self.generate_file_list_from_packages(self.dependencies(packages_to_deploy))
             if '/lib/libpthread-2.31.so' in fs_:
                 dfsfdf = 1
             if '/lib/libpthread-2.31.so' in file_list:
                 dfsfdf = 1
-            if '/usr/lib64/python3.8/lib-dynload/unicodedata.cpython-38-x86_64-linux-gnu.so' not in file_list:
+            if 'java-11' in file_list:
                 fdsfsdfds=1
             file_list.extend(fs_)
 
@@ -1684,7 +1770,7 @@ terrarium_assembler --debug --stage-pack=./out-debug "%(specfile_)s" --stage-mak
                 if 'libpthread-2.31.so' in f:
                     wtff = 1
 
-                if '/usr/bin/ld' in f:
+                if 'java-11' in f:
                     wtff = 1
 
                 if self.br.is_need_patch(f):  
@@ -1741,26 +1827,39 @@ terrarium_assembler --debug --stage-pack=./out-debug "%(specfile_)s" --stage-mak
                             self.add(f, libfile, recursive=False)
                             # shutil.copy2(f, os.path.join(root_dir, libfile))
                             # add(f, arcname=libfile, recursive=False)  
+                pass
+                # if os.path.exists('/home/stas/projects/deploy-for-audit/linux_distro/out/lib64/jvm/java-11-openjdk-11.0.11.0.9-4.fc33.x86_64-slowdebug/lib/modules'):
+                #     fsdfsdf = 1  
+
         
-        
+            dfsfsdfsdf=1
+            with open('file-list-from-packages.txt', 'w', encoding='utf-8') as lf:
+                lf.write('\n'.join(file_list))
+
+            self.cmd('sudo chmod a+r /usr/lib/cups -R')
             for f in file_list:
                 copy_file_to_environment(f)
 
+            # if os.path.exists('/home/stas/projects/deploy-for-audit/linux_distro/out/lib64/jvm/java-11-openjdk-11.0.11.0.9-4.fc33.x86_64-slowdebug/lib/modules'):
+            #     fsdfsdf = 1  
+
             if self.fs:    
                 for folder_ in self.fs.folders:
                     for dirpath, dirnames, filenames in os.walk(folder_):
                         for filename in filenames:
                             f = os.path.join(dirpath, filename)
-                            if self.br.is_need_exclude(f):
-                                continue
+                            # if self.br.is_need_exclude(f):
+                            #     continue
                             # if not self.should_copy(f):
                             #     continue
                             if self.br.is_need_patch(f):  
                                 self.process_binary(f)
                                 continue
                             libfile = os.path.join(self.root_dir, f.replace(folder_, 'pbin'))
-                            if self.br.is_need_exclude(libfile):
-                                continue
+                            if 'java-11' in libfile:
+                                erwerew =1
+                            # if self.br.is_need_exclude(libfile):
+                            #     continue
                             self.add(f, libfile, recursive=False)
                     pass
 
@@ -1796,7 +1895,12 @@ terrarium_assembler --debug --stage-pack=./out-debug "%(specfile_)s" --stage-mak
                         lf.write(f'   {pat_} \n')
 
             # size_ = sum(file.stat().st_size for file in pathlib.Path(self.root_dir).rglob('*'))
+            
+            # Postprocessing, removing not needed files after installing python modules, etc
+            self.remove_exclusions()
+            
             size_  = folder_size(self.root_dir, follow_symlinks=False)
+
             print("Size ", size_/1024/1024, 'Mb')
 
         if self.args.stage_make_isoexe:
@@ -1807,15 +1911,20 @@ terrarium_assembler --debug --stage-pack=./out-debug "%(specfile_)s" --stage-mak
                 label = self.spec.label
             installscript = "install-me.sh" % vars()
             installscriptpath = os.path.abspath(os.path.join("tmp/", installscript))
-            print("*"*10)
-            print(self.curdir)
-            print("*"*10)
+            if os.path.exists(installscriptpath):
+                os.unlink(installscriptpath)
+
+            pmode = ''
+            if shutil.which('pbzip2'):
+                pmode = ' --threads 8 --pbzip2 ' 
             os.chdir(self.curdir)
             self.cmd(f'chmod a+x {root_dir}/install-me')
-            scmd = ('''
-            makeself.sh --needroot %(root_dir)s  %(installscriptpath)s "Installation" ./install-me             
+            scmd = (f'''
+            makeself.sh {pmode} --needroot {root_dir} {installscriptpath} "Installation" ./install-me             
         ''' % vars()).replace('\n', ' ').strip()
-            self.cmd(scmd)
+            if not self.cmd(scmd)==0:
+                print(f'« {scmd} » failed!')
+                return
 
             filename = "%(time_prefix)s-%(label)s-dm.iso" % vars()
             isodir = root_dir + '.iso'
@@ -1826,5 +1935,8 @@ terrarium_assembler --debug --stage-pack=./out-debug "%(specfile_)s" --stage-mak
         ''' % vars()).replace('\n', ' ').strip()
             print(scmd)
             os.system(scmd)
+            os.chdir(isodir)
+            scmd = f'''ln -sf {filename} last.iso'''
+            self.cmd(scmd)
             print(filepath)
     
